/*
 * jcmainct.c
 *
 * Copyright (C) 1994, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains the main buffer controller for compression.
 * The main buffer lies between the pre-processor and the JPEG
 * compressor proper; it holds downsampled data in the JPEG colorspace.
 */

#include "jinclude.h"
#include "jpegint.h"


/* Private buffer controller object */

typedef struct {
  struct jpeg_c_main_controller pub; /* public fields */

  /* If using full-image storage, this array holds pointers to virtual-array
   * control blocks for each component.  Unused if not full-image storage.
   */
  jvirt_sarray_ptr virtual[MAX_COMPONENTS];
  /* If using just a strip buffer, this points to the entire set of buffers
   * (we allocate one for each component).  In the virtual-array case, this
   * points to the currently accessible strips of the virtual arrays.
   */
  JSAMPARRAY buffer[MAX_COMPONENTS];
  /* This vector points to the current strips of the components being written
   * during the current scan.
   */
  JSAMPARRAY cur_buffer[MAX_COMPS_IN_SCAN];

  JDIMENSION rowgroup_ctr;	/* counts row groups received in MCU row */
  JDIMENSION mcu_ctr;		/* counts MCUs output from current row */
  boolean suspended;		/* remember if we suspended output */
} my_main_controller;

typedef my_main_controller * my_main_ptr;


/*
 * Initialize for a processing pass.
 */

METHODDEF void
start_pass (j_compress_ptr cinfo, J_BUF_MODE pass_mode)
{
  my_main_ptr main = (my_main_ptr) cinfo->main;
  int ci, cindex;

  switch (pass_mode) {
  case JBUF_PASS_THRU:
    if (main->virtual[0] != NULL)
      ERREXIT(cinfo, JERR_BAD_BUFFER_MODE);
    for (ci = 0; ci < cinfo->comps_in_scan; ci++) {
      cindex = cinfo->cur_comp_info[ci]->component_index;
      main->cur_buffer[ci] = main->buffer[cindex];
    }
    main->rowgroup_ctr = 0;
    main->mcu_ctr = 0;
    main->suspended = FALSE;
    break;
  default:
    ERREXIT(cinfo, JERR_BAD_BUFFER_MODE);
    break;
  }
}


/*
 * Process some data.
 */

METHODDEF void
process_data (j_compress_ptr cinfo,
	      JSAMPARRAY input_buf, JDIMENSION *in_row_ctr,
	      JDIMENSION in_rows_avail)
{
  my_main_ptr main = (my_main_ptr) cinfo->main;

  while (*in_row_ctr < in_rows_avail) {
    /* Read input data if we haven't filled the main buffer yet */
    if (main->rowgroup_ctr < DCTSIZE)
      (*cinfo->prep->pre_process_data) (cinfo,
					input_buf, in_row_ctr, in_rows_avail,
					main->buffer, &main->rowgroup_ctr,
					(JDIMENSION) DCTSIZE);

    /* If we have a full MCU row buffered, emit data. */
    /* Note that preprocessor ensures this will be true at bottom of image. */
    if (main->rowgroup_ctr == DCTSIZE) {
      (*cinfo->coef->compress_data) (cinfo, main->buffer, &main->mcu_ctr);
      /* If compressor did not consume the whole row, then we must need to
       * suspend processing and return to the application.  In this situation
       * we pretend we didn't yet consume the last input row; otherwise, if
       * it happened to be the last row of the image, the application would
       * think we were done.
       */
      if (main->mcu_ctr < cinfo->MCUs_per_row) {
	if (! main->suspended) {
	  (*in_row_ctr)--;
	  main->suspended = TRUE;
	}
	return;
      }
      /* We did finish the row.  Undo our little suspension hack if a previous
       * call suspended; then mark the main buffer empty.
       */
      if (main->suspended) {
	(*in_row_ctr)++;
	main->suspended = FALSE;
      }
      main->mcu_ctr = 0;
      main->rowgroup_ctr = 0;
    }
  }
}


/*
 * Clean up at the end of a pass.
 */

METHODDEF void
finish_pass (j_compress_ptr cinfo)
{
  /* No work for myself, but pass it on to subsidiary objects */
  (*cinfo->prep->finish_pass) (cinfo);
  (*cinfo->coef->finish_pass) (cinfo);
}


/*
 * Initialize main buffer controller.
 */

GLOBAL void
jinit_c_main_controller (j_compress_ptr cinfo, boolean need_full_buffer)
{
  my_main_ptr main;
  int ci;
  jpeg_component_info *compptr;

  main = (my_main_ptr)
    (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_IMAGE,
				SIZEOF(my_main_controller));
  cinfo->main = (struct jpeg_c_main_controller *) main;
  main->pub.start_pass = start_pass;
  main->pub.process_data = process_data;
  main->pub.finish_pass = finish_pass;

  /* Create the buffer.  It holds downsampled data, so each component
   * may be of a different size.
   */
  if (need_full_buffer) {
    /* Allocate a full-image virtual array for each component */
    for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
	 ci++, compptr++) {
      main->virtual[ci] = (*cinfo->mem->request_virt_sarray)
	((j_common_ptr) cinfo, JPOOL_IMAGE,
	 compptr->width_in_blocks * DCTSIZE,
	 compptr->height_in_blocks * DCTSIZE,
	 (JDIMENSION) (compptr->v_samp_factor * DCTSIZE));
    }
  } else {
    main->virtual[0] = NULL;	/* flag for no virtual arrays */
    /* Allocate a strip buffer for each component */
    for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
	 ci++, compptr++) {
      main->buffer[ci] = (*cinfo->mem->alloc_sarray)
	((j_common_ptr) cinfo, JPOOL_IMAGE,
	 compptr->width_in_blocks * DCTSIZE,
	 (JDIMENSION) (compptr->v_samp_factor * DCTSIZE));
    }
  }
}
