/*
 * jddctmgr.c
 *
 * Copyright (C) 1994, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains the inverse-DCT management logic.
 * This code selects a particular IDCT implementation to be used,
 * and it performs related housekeeping chores.
 */

#include "jinclude.h"
#include "jpegint.h"


/* Extern declarations for the core IDCT routines.  They are kept in
 * separate files to ease machine-dependent tuning (eg, assembly coding).
 *
 * The IDCT routine is given a pointer to the input JBLOCK and a pointer
 * to an output work area.  The output area is declared int, not JSAMPLE,
 * because the outputs are expected to be signed (range +-CENTERJSAMPLE)
 * and because we must allow for slightly out-of-range outputs due to
 * quantization and roundoff error.  This module is responsible for doing
 * range-limiting, changing to unsigned representation, and storing the
 * final results into the sample arrays.
 */

EXTERN void j_rev_dct JPP((JCOEFPTR input, int * output));


/* Private subobject for this module */

typedef struct {
  struct jpeg_inverse_dct pub;	/* public fields */
  /* no private fields for now */
} my_idct_controller;

typedef my_idct_controller * my_idct_ptr;


/*
 * Initialize for a processing pass.
 * Verify that all referenced Q-tables are present, and set up
 * the multiplier tables for each one.
 */

METHODDEF void
start_pass (j_decompress_ptr cinfo)
{
  int ci, qtblno, i;
  JDQUANT_TBL * qtbl;

  for (ci = 0; ci < cinfo->comps_in_scan; ci++) {
    qtblno = cinfo->cur_comp_info[ci]->quant_tbl_no;
    /* Make sure specified quantization table is present */
    if (qtblno < 0 || qtblno >= NUM_QUANT_TBLS ||
	cinfo->quant_tbl_ptrs[qtblno] == NULL)
      ERREXIT1(cinfo, JERR_NO_QUANT_TABLE, qtblno);
    /* Compute multipliers for this quant table */
    /* We may do this more than once for same table, but it's not a big deal */
    qtbl = (JDQUANT_TBL *) cinfo->quant_tbl_ptrs[qtblno];
    for (i = 0; i < DCTSIZE2; i++) {
      /* For now, multipliers are same as raw quantization coefficients */
      qtbl->multipliers[i] = (MULTIPLIER) qtbl->pub.quantval[i];
    }
  }
}


/*
 * Perform inverse DCT on one or more blocks of a component.
 *
 * The coefficients are passed in coef_blocks[]; the output samples
 * go into the sample_data[] array starting at position start_row/start_col,
 * and moving to the right for any additional blocks.  Note that the
 * output data for each block is of size DCT_scaled_size*DCT_scaled_size.
 * NB: AS IS, CODE ASSUMES DCT_scaled_size == DCTSIZE.
 */

METHODDEF void
inverse_DCT (j_decompress_ptr cinfo, jpeg_component_info * compptr,
	     JBLOCKROW coef_blocks, JSAMPARRAY sample_data,
	     JDIMENSION start_row, JDIMENSION start_col,
	     JDIMENSION num_blocks)
{
  int workspace[DCTSIZE2];	/* output area for IDCT subroutine */
  register int *workspaceptr;
  JDIMENSION bi;
  register JSAMPROW elemptr;
  register JSAMPLE *range_limit = cinfo->sample_range_limit + CENTERJSAMPLE;
#if DCTSIZE != 8
  register int elemc;
#endif
  register int elemr;

  sample_data += start_row;	/* fold in the vertical offset once */

  for (bi = 0; bi < num_blocks; bi++, start_col += DCTSIZE) {
    /* Perform inverse DCT */
    j_rev_dct(coef_blocks[bi], workspace);

    /* Output the data into the sample array.
     * Note change from signed to unsigned representation:
     * DCT calculation works with values +-CENTERJSAMPLE,
     * but sample arrays always hold 0..MAXJSAMPLE.
     * We have to do range-limiting because of quantization errors in the
     * DCT/IDCT phase.  We use the sample_range_limit[] table to do this
     * quickly; the CENTERJSAMPLE offset is folded into table indexing.
     */
    workspaceptr = workspace;

    for (elemr = 0; elemr < DCTSIZE; elemr++) {
      elemptr = sample_data[elemr] + start_col;
#if DCTSIZE == 8		/* unroll the inner loop */
      *elemptr++ = range_limit[*workspaceptr++];
      *elemptr++ = range_limit[*workspaceptr++];
      *elemptr++ = range_limit[*workspaceptr++];
      *elemptr++ = range_limit[*workspaceptr++];
      *elemptr++ = range_limit[*workspaceptr++];
      *elemptr++ = range_limit[*workspaceptr++];
      *elemptr++ = range_limit[*workspaceptr++];
      *elemptr++ = range_limit[*workspaceptr++];
#else
      for (elemc = DCTSIZE; elemc > 0; elemc--) {
	*elemptr++ = range_limit[*workspaceptr++];
      }
#endif
    }
  }
}


/*
 * Clean up at the end of a pass.
 */

METHODDEF void
finish_pass (j_decompress_ptr cinfo)
{
  /* No work */
}


/*
 * Initialize IDCT manager.
 */

GLOBAL void
jinit_inverse_dct (j_decompress_ptr cinfo)
{
  my_idct_ptr idct;

  idct = (my_idct_ptr)
    (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_IMAGE,
				SIZEOF(my_idct_controller));
  cinfo->idct = (struct jpeg_inverse_dct *) idct;

  idct->pub.start_pass = start_pass;
  idct->pub.inverse_DCT = inverse_DCT;
  idct->pub.finish_pass = finish_pass;
}
